///////////////////////////////////////////////////////////
/////////////Classe de Volume Rendering : VolRender////////
//////////////////Cyril Crassin 12 Octobre 2003////////////
///////////////////////////////////////////////////////////

#include "VolRender.h"

#include <stdio.h>
#include <math.h>

#include <GL/glut.h>

GLUX_LOAD(GL_EXT_texture3D);
GLUX_LOAD(GL_ARB_multitexture);

//Ya un truc bizard sous linux, les extensions passes ds le standard ne semblent plus exister...
#ifndef WIN32
#define glActiveTextureARB glActiveTexture
#define glTexSubImage3DEXT glTexSubImage3D
#endif

extern int ttt;
/////////Cg/////////////
void VolRender::CheckCgError() {
	CGerror err = cgGetError();

	if (err != CG_NO_ERROR) {
	 throw CgError(cgGetErrorString(err), cgGetLastListing(shaderContext));

//	 exit(1);
   }
}
int VolRender::initCg(){
	// Cree le contexte
	shaderContext = cgCreateContext();

	///////////Vertex
	if( cgGLIsProfileSupported(CG_PROFILE_ARBVP1) )
        vertexProfile = CG_PROFILE_ARBVP1;
    else if( cgGLIsProfileSupported(CG_PROFILE_VP30) )
        vertexProfile = CG_PROFILE_VP30;
	else if( cgGLIsProfileSupported(CG_PROFILE_VP20) )
        vertexProfile = CG_PROFILE_VP20;
    else {
		throw CgError("Impossible d'initialiser le Vertex Shader", "Extensions non supportees");
    }

	char *args[2];
	args[0]=new char[8];
	args[1]=0;
	strcpy(args[0],"-posinv");

	// Cree le vertex program
	vertexProgram = cgCreateProgramFromFile(shaderContext, CG_SOURCE, "volrendVert.cg", vertexProfile, NULL, (const char **)args);
	CheckCgError();

	// Charge le Vertex Program
	cgGLLoadProgram(vertexProgram);
	CheckCgError();


	///////////Fragment
	if (cgGLIsProfileSupported(CG_PROFILE_FP30))
        fragmentProfile = CG_PROFILE_FP30;
	else if (cgGLIsProfileSupported(CG_PROFILE_FP20))
        fragmentProfile = CG_PROFILE_FP20;
    else if (cgGLIsProfileSupported(CG_PROFILE_ARBFP1)){
        fragmentProfile = CG_PROFILE_ARBFP1;
	} else {
        throw CgError("Fragment programming extensions (GL_ARB_fragment_program or GL_NV_fragment_program) not supported", "Extensions non supportees");
    }

	


	CheckCgError();

	////Fragment prog////
	fragmentProgram = cgCreateProgramFromFile(shaderContext, CG_SOURCE, "volrendFrag.cg", fragmentProfile, NULL, NULL);

    cgGLLoadProgram(fragmentProgram);

	CheckCgError();

	cgParam_lightPos = cgGetNamedParameter(fragmentProgram, "lightPos");


	cgGLBindProgram( vertexProgram );
	cgGLEnableProfile( vertexProfile );

	cgGLBindProgram(fragmentProgram);
	cgGLEnableProfile(fragmentProfile);

	return 0;
}


//Initialisation des extensions
int VolRender::initExts(){
/*	int ext_texture3D =0;

	ext_texture3D = (strstr((const char*)glGetString(GL_EXTENSIONS), "GL_EXT_texture3D") != 0);

	if (!ext_texture3D) {
		printf("\nThis demo requires the 3D texture extension to OpenGL.\n");
		return 1;
    }

	glTexImage3DEXT=(PFNGLTEXIMAGE3DEXTPROC)wglGetProcAddress("glTexImage3DEXT");
	glTexSubImage3DEXT=(PFNGLTEXSUBIMAGE3DEXTPROC)wglGetProcAddress("glTexSubImage3DEXT");

	glActiveTextureARB   = (PFNGLACTIVETEXTUREARBPROC) wglGetProcAddress("glActiveTextureARB");

	glMultiTexCoord1fARB = (PFNGLMULTITEXCOORD1FARBPROC) wglGetProcAddress("glMultiTexCoord1fARB");
	glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC) wglGetProcAddress("glMultiTexCoord2fARB");
	glMultiTexCoord3fARB = (PFNGLMULTITEXCOORD3FARBPROC) wglGetProcAddress("glMultiTexCoord3fARB");
	glMultiTexCoord4fARB = (PFNGLMULTITEXCOORD4FARBPROC) wglGetProcAddress("glMultiTexCoord4fARB");
*/
	 gluxInit();


	return 0;
}


int VolRender::init(){

  initExts();
 
  static GLfloat splane[4] = {1.f/200.f, 0.f, 0.f, .5f};
  static GLfloat rplane[4] = {0, 1.f/200.f, 0, .5f};
  static GLfloat tplane[4] = {0, 0, 1.f/200.f, .5f};
  

  glGenTextures(1, &texID);
   
  //////////
  glActiveTextureARB(GL_TEXTURE0_ARB);
  
  glBindTexture(GL_TEXTURE_3D_EXT, texID);
  
  glEnable(GL_TEXTURE_3D_EXT);
	
 

	////////
/*	glGenTextures(1, &gradID);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glBindTexture(GL_TEXTURE_3D_EXT, gradID);
	glEnable(GL_TEXTURE_3D_EXT);

    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);

    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);

    glTexGenfv(GL_S, GL_OBJECT_PLANE, splane);
    glTexGenfv(GL_T, GL_OBJECT_PLANE, tplane);
    glTexGenfv(GL_R, GL_OBJECT_PLANE, rplane);
*/
	//////////////


	//
    glEnable(GL_CLIP_PLANE0);
    glEnable(GL_CLIP_PLANE1);
    glEnable(GL_CLIP_PLANE2);
    glEnable(GL_CLIP_PLANE3);
    glEnable(GL_CLIP_PLANE4);
    glEnable(GL_CLIP_PLANE5);


    glGenTextures(1, &tableID);

    glBindTexture(GL_TEXTURE_2D, tableID);
    

    
    //Generation de la palette
    float tbl[256*4];
    
    for(int i=0; i<256; i++){
      
      tbl[0+i*4]=cos(i/10.0);
      tbl[1+i*4]=cos(i/10.0)*sin(i/100.0);
      tbl[2+i*4]=sin(i/40.0);
      tbl[3+i*4]=i/255.0;
      
      /*	tbl[0+i*4]=0;
		tbl[1+i*4]=0;
		tbl[2+i*4]=cos(i/200.0);
		tbl[3+i*4]=i/255.0;*/
      

      /*tbl[i*4]=0.5;
	tbl[1+i*4]=0.5;
	tbl[2+i*4]=1.0;
	if(i==0)
	tbl[3+i*4]=i/255.0;
	else
	tbl[3+i*4]=1.0;*/
      
      /*tbl[i*4]=i/255.0;
	tbl[1+i*4]=i/255.0;
	tbl[2+i*4]=i/255.0;
	tbl[3+i*4]=i/255.0;*/
    }
    
	setPalette(tbl);


    initCg();
    
    
    planesList=glGenLists(1);
    
    buildPlanes();
    
    return 0;
}

void VolRender::setPalette(float *pal){

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 256, 1, 0, GL_RGBA , GL_FLOAT, pal);
    //gluBuild2DMipmaps(GL_TEXTURE_2D,  GL_RGBA, 256, 1,  GL_RGBA , GL_UNSIGNED_BYTE, tbl);
    
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

}

//Chargement de fichiers RAW, methode rapide
void VolRender::loadRawData(char *fich, IVector size, int nbcomp){
	fieldSize=size;

	int tx=size.x, ty=size.y, tz=size.z;

	if(tex3ddata && my3ddata)
		delete [] tex3ddata;

	my3ddata=1;

	tex3ddata=new float[(tx+0)*(ty+0)*(tz+0)*nbcomp];

	FILE *file=0;
	file=fopen(fich, "r");

	if(!file)
		throw Error("LoadRawData: Erreur, impossible de charger le fichier raw");

	unsigned char *buff=new unsigned char[tx*ty*tz*nbcomp];

/*	for(int cc=0; cc<(tx+2)*(ty+2)*(tz+2)*nbcomp; cc++)
		tex3ddata[cc]=0;
*/
	fread(buff, 1, tx*ty*tz*nbcomp, file);

	for(int k=0; k<tz; k++)
		for(int j=0; j<ty; j++)
			for(int i=0; i<tx; i++)
				for(int cmp=0; cmp<nbcomp; cmp++){
				
					if(i<1 || j<1 || k<1 || i>(tx-2) || j>(ty-2) || k>(tz-2))
						tex3ddata[cmp + (i+0)*nbcomp + (j+0)*nbcomp*tx + (k+0)*nbcomp*tx*ty]=buff[cmp + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]/255.0;
					else
						tex3ddata[cmp + (i+0)*nbcomp + (j+0)*nbcomp*tx + (k+0)*nbcomp*tx*ty]=(buff[cmp + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty] +buff[cmp + (i-1)*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]+buff[cmp + (i+1)*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]+buff[cmp + i*nbcomp + (j-1)*nbcomp*tx + k*nbcomp*tx*ty]+buff[cmp + i*nbcomp + (j+1)*nbcomp*tx + k*nbcomp*tx*ty]+buff[cmp + i*nbcomp + j*nbcomp*tx + (k-1)*nbcomp*tx*ty]+buff[cmp + i*nbcomp + j*nbcomp*tx + (k+1)*nbcomp*tx*ty])/(255.0*7.0);
					//adoucissement
				}
			


	fclose(file);


/*	fieldSize.x=128; fieldSize.y=128; fieldSize.z=128;

	delete [] tex3ddata;
	tex3ddata=new float[128*128*128*4];


	for(int i=0; i<128*128*128*4; i++)
		tex3ddata[i]=i%256;*/

	updateDataTexture();
	//my3ddata=0;
}

//Chargement 2 a fusioner et generaliser
void VolRender::loadRawDataPart(char *fich, IVector size, int nbcomp, IVector partPos, IVector partSize, IVector fsize){

	if(fsize.x!=-1 || nbcomp!=nbComp || !tex3ddata || !my3ddata) {
		if(tex3ddata && my3ddata)
			delete [] tex3ddata;

		my3ddata=1;
		nbComp=nbcomp;

		if(fsize.x!=-1)
			fieldSize=fsize;

		tex3ddata=new float[(fieldSize.x+0)*(fieldSize.y+0)*(fieldSize.z+0)*nbComp];

	}

	int tx=fieldSize.x, ty=fieldSize.y, tz=fieldSize.z;


	FILE *file;
	file=fopen(fich, "r");


	int oriTx=size.x, oriTy=size.y, oriTz=size.z;
	//int oriTx=1024, oriTy=1024, oriTz=741;
	float zoomX=(float)partSize.x/(float)tx, 
		zoomY=(float)partSize.y/(float)ty, 
		zoomZ=(float)partSize.z/(float)tz;

	unsigned char comp;

	printf("%f %f %f\n", zoomX, zoomY, zoomZ);

	for(int k=0; k<tz; k++)
		for(int j=0; j<ty; j++)
			for(int i=0; i<tx; i++)
				for(int cmp=0; cmp<nbcomp; cmp++){
				//	p.x=partPos.x+
					fseek(file, (int)(i*zoomX+partPos.x) + (int)(j*zoomY+partPos.y )*oriTx + (int)(k*zoomZ+partPos.z)*oriTx*oriTy, 0);
					fread(&comp, 1, 1, file);
					tex3ddata[cmp + (i)*nbcomp + (j)*nbcomp*tx + (k)*nbcomp*tx*ty]=comp/255.0;
				}
			


	fclose(file);

	updateDataTexture();

//	my3ddata=0;
}

void VolRender::updateDataTexture(){
	int tx=fieldSize.x, ty=fieldSize.y, tz=fieldSize.z;

	glBindTexture(GL_TEXTURE_3D_EXT, texID);

	//glPixelTransferf(GL_ALPHA_SCALE,2.0f);


	//gluBuild3DMipmaps(GL_TEXTURE_3D_EXT, GL_ALPHA, tx, ty, tz, GL_COLOR_INDEX, GL_FLOAT, sf);

	if(nbComp==1){
		gluBuild3DMipmaps(GL_TEXTURE_3D_EXT, 4, tx, ty, tz, GL_ALPHA, GL_FLOAT, tex3ddata);
	//	glTexImage3DEXT(GL_TEXTURE_3D_EXT, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, tx, ty, tz, 0, GL_ALPHA , GL_FLOAT, tex3ddata);
	//	glTexImage3DEXT(GL_TEXTURE_3D_EXT, 0, GL_RGBA, tx, ty, tz, 0, GL_ALPHA , GL_FLOAT, tex3ddata);
	}else{
		gluBuild3DMipmapLevels(GL_TEXTURE_3D_EXT, GL_RGBA, tx, ty, tz, GL_RGBA, GL_FLOAT, 0, 0, 0, tex3ddata);
	}


	//Clamping
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_CLAMP_TO_EDGE );

    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	/*int col[]={0,0,0,0};
	float p=1;

	glTexParameterfv(GL_TEXTURE_3D_EXT, GL_TEXTURE_PRIORITY, &p);*/

}

void VolRender::updateDataTextureRegion(float *data, IVector pos, IVector size){

	glBindTexture(GL_TEXTURE_3D_EXT, texID);

	glTexSubImage3DEXT(GL_TEXTURE_3D_EXT, 0, pos.x, pos.y, pos.z, size.x, size.y, size.z, GL_ALPHA, GL_FLOAT, data);

}

void VolRender::bindScalarField(float *sf, IVector size){
	fieldSize=size;

	if(tex3ddata && my3ddata)
		delete [] tex3ddata;

	tex3ddata=sf;
	my3ddata=0;

	nbComp=1;

	updateDataTexture();

	//buildPlanes();
}

void VolRender::bindRGBAField(float *sf, IVector size){
	fieldSize=size;

	int tx=size.x, ty=size.y, tz=size.z;
	nbComp=4;

	if(tex3ddata && my3ddata)
		delete [] tex3ddata;

	tex3ddata= sf;

	updateDataTexture();

	buildPlanes();
}

void VolRender::draw0(){

/*	ttt=glutGet(GLUT_ELAPSED_TIME);
	printf("1: 0\n");*/

	static GLfloat splane[4] = {1.f/200.f, 0.f, 0.f, .5f};
    static GLfloat rplane[4] = {0, 1.f/200.f, 0, .5f};
    static GLfloat tplane[4] = {0, 0, 1.f/200.f, .5f};



	glActiveTextureARB(GL_TEXTURE0_ARB);
	glBindTexture(GL_TEXTURE_3D_EXT, texID);

	glEnable(GL_TEXTURE_3D_EXT);

	/*cgGLEnableProfile( vertexProfile );
	cgGLEnableProfile(fragmentProfile);*/


	glEnable(GL_CLIP_PLANE0);
    glEnable(GL_CLIP_PLANE1);
    glEnable(GL_CLIP_PLANE2);
    glEnable(GL_CLIP_PLANE3);
    glEnable(GL_CLIP_PLANE4);
    glEnable(GL_CLIP_PLANE5);


///////////////////////dd/////////////////

	GLfloat offS, offT, offR; // mapping texture to planes 

	GLdouble clipplane0[] = {-1,  0,  0, 100};
	GLdouble clipplane1[] = { 1,  0,  0, 100};
	GLdouble clipplane2[] = { 0, -1,  0, 100};
	GLdouble clipplane3[] = { 0,  1,  0, 100};
	GLdouble clipplane4[] = { 0,  0, -1, 100};
	GLdouble clipplane5[] = { 0,  0,  1, 100};

	// define a cutting plane
	GLdouble cutplane[] = {0.f, -.5f, -2.f, 50.f};


    offS = 100.f/fieldSize.x;
    offT = 100.f/fieldSize.y;
    offR = 100.f/fieldSize.z;
    
    clipplane0[W] = 90.f - offS;
    clipplane1[W] = 90.f - offS;
    clipplane2[W] = 90.f - offT;
    clipplane3[W] = 90.f - offT;
    clipplane4[W] = 90.f - offR;
    clipplane5[W] = 90.f - offR;


	///////Perspective////////
//	glScalef(0.010,0.010,0.010);
	glTranslatef(0,0,-250);
	//////////////////////////

//	printf("2: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

	 // GL_MODELVIEW
    if(cut) {
		cutplane[W] = cutbias;
		glClipPlane(GL_CLIP_PLANE5, cutplane);
    }

	glColor4f(1,1,1,1);
	

    glPushMatrix(); // identity
	
		//pour le fluide 2D
	//	glScalef(1.0,1.0,fieldSize.z/150.0);

	//	glRotatef(objangle[X], 0.f, 1.f, 0.f);//0 1 0
		glRotatef(objangle[X], 0.f, 0.f, 1.f);
		glRotatef(objangle[Y], -1.f, 0.f, 0.f);



		glClipPlane(GL_CLIP_PLANE0, clipplane0);
		glClipPlane(GL_CLIP_PLANE1, clipplane1);
		glClipPlane(GL_CLIP_PLANE2, clipplane2);
		glClipPlane(GL_CLIP_PLANE3, clipplane3);
		glClipPlane(GL_CLIP_PLANE4, clipplane4);
		if(!cut)
			glClipPlane(GL_CLIP_PLANE5, clipplane5);

    glPopMatrix(); // back to identity
	
	
//	printf("3: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

	glActiveTextureARB(GL_TEXTURE2_ARB);
	glDisable(GL_TEXTURE_2D);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_3D_EXT);



//	printf("4: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

/*	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tableID);*/



	//glScalef( 3,  3, 3);

    glMatrixMode(GL_TEXTURE);
    glPushMatrix(); // identity 
		glTranslatef( .5f,  .5f, .5f);
	//	glRotatef(objangle[X], 0.f, -1.f, 0.f);
	
		glRotatef(objangle[Y], -1.f, 0.f, 0.f);
		glRotatef(objangle[X], 0.f, 0.f, -1.0f);
		

		//glScalef(objangle[X]/10.0, objangle[X]/10.0, objangle[X]/10.0);
		glTranslatef( -.5f,  -.5f, -.5f);

		

		//glScalef( 0.3,  0.3, 0.3);


		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		//glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	
		
		glCallList(planesList);

		glDisable(GL_TEXTURE_3D_EXT);
		//glDisable(GL_BLEND);

	glPopMatrix(); // back to identity

    glMatrixMode(GL_MODELVIEW);

//	printf("5: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

	//raz
	cgGLDisableProfile( vertexProfile );
	cgGLDisableProfile(fragmentProfile);

	glActiveTextureARB(GL_TEXTURE2_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_3D_EXT);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glDisable(GL_TEXTURE_3D_EXT);
	glDisable(GL_TEXTURE_2D);

	
    glDisable(GL_CLIP_PLANE0);
    glDisable(GL_CLIP_PLANE1);
    glDisable(GL_CLIP_PLANE2);
    glDisable(GL_CLIP_PLANE3);
    glDisable(GL_CLIP_PLANE4);
    glDisable(GL_CLIP_PLANE5);


//	printf("6: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);
	/*glPushMatrix();
		glScalef(90,90,90);
		glRotatef(objangle[X], 0.f, 0.f, 1.f);
		glRotatef(objangle[Y], -1.f, 0.f, 0.f);

		drawCube();
	glPopMatrix();*/




}

void VolRender::draw(){

/*	ttt=glutGet(GLUT_ELAPSED_TIME);
	printf("1: 0\n");*/

//	updateDataTexture();  //pert 1FPS

	static GLfloat splane[4] = {1.f/200.f, 0.f, 0.f, .5f};
    static GLfloat rplane[4] = {0, 1.f/200.f, 0, .5f};
    static GLfloat tplane[4] = {0, 0, 1.f/200.f, .5f};



	glActiveTextureARB(GL_TEXTURE0_ARB);
	glBindTexture(GL_TEXTURE_3D_EXT, texID);

	glEnable(GL_TEXTURE_3D_EXT);

	cgGLEnableProfile( vertexProfile );
	cgGLEnableProfile(fragmentProfile);


	glEnable(GL_CLIP_PLANE0);
    glEnable(GL_CLIP_PLANE1);
    glEnable(GL_CLIP_PLANE2);
    glEnable(GL_CLIP_PLANE3);
    glEnable(GL_CLIP_PLANE4);
    glEnable(GL_CLIP_PLANE5);


///////////////////////dd/////////////////

	GLfloat offS, offT, offR; // mapping texture to planes 

	GLdouble clipplane0[] = {-1,  0,  0, 100};
	GLdouble clipplane1[] = { 1,  0,  0, 100};
	GLdouble clipplane2[] = { 0, -1,  0, 100};
	GLdouble clipplane3[] = { 0,  1,  0, 100};
	GLdouble clipplane4[] = { 0,  0, -1, 100};
	GLdouble clipplane5[] = { 0,  0,  1, 100};

	// define a cutting plane
	GLdouble cutplane[] = {0.f, -.5f, -2.f, 50.f};


    offS = 100.f/fieldSize.x;
    offT = 100.f/fieldSize.y;
    offR = 100.f/fieldSize.z;
    
    clipplane0[W] = 90.f - offS;
    clipplane1[W] = 90.f - offS;
    clipplane2[W] = 90.f - offT;
    clipplane3[W] = 90.f - offT;
    clipplane4[W] = 90.f - offR;
    clipplane5[W] = 90.f - offR;


	///////Perspective////////
//	glScalef(0.010,0.010,0.010);
	glTranslatef(0,0,-250);
	//////////////////////////

//	printf("2: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

	 // GL_MODELVIEW
    if(cut) {
		cutplane[W] = cutbias;
		glClipPlane(GL_CLIP_PLANE5, cutplane);
    }

	glColor4f(1,1,1,1);
	

    glPushMatrix(); // identity
	
		//pour le fluide 2D
	//	glScalef(1.0,1.0,fieldSize.z/150.0);

	//	glRotatef(objangle[X], 0.f, 1.f, 0.f);//0 1 0
		glRotatef(objangle[X], 0.f, 0.f, 1.f);
		glRotatef(objangle[Y], -1.f, 0.f, 0.f);



		glClipPlane(GL_CLIP_PLANE0, clipplane0);
		glClipPlane(GL_CLIP_PLANE1, clipplane1);
		glClipPlane(GL_CLIP_PLANE2, clipplane2);
		glClipPlane(GL_CLIP_PLANE3, clipplane3);
		glClipPlane(GL_CLIP_PLANE4, clipplane4);
		if(!cut)
			glClipPlane(GL_CLIP_PLANE5, clipplane5);

    glPopMatrix(); // back to identity
	
	
//	printf("3: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

	glActiveTextureARB(GL_TEXTURE2_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tableID);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_3D_EXT);
	glBindTexture(GL_TEXTURE_3D_EXT, gradID);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_3D_EXT);
//	glBindTexture(GL_TEXTURE_3D_EXT, texID);



//	printf("4: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

/*	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tableID);*/



	//glScalef( 3,  3, 3);

    glMatrixMode(GL_TEXTURE);
    glPushMatrix(); // identity 
		glTranslatef( .5f,  .5f, .5f);
	//	glRotatef(objangle[X], 0.f, -1.f, 0.f);
	
		glRotatef(objangle[Y], -1.f, 0.f, 0.f);
		glRotatef(objangle[X], 0.f, 0.f, -1.f);
		

		//glScalef(objangle[X]/10.0, objangle[X]/10.0, objangle[X]/10.0);
		glTranslatef( -.5f,  -.5f, -.5f);

		

		//glScalef( 0.3,  0.3, 0.3);


		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		//glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	
		
		glCallList(planesList);

		glDisable(GL_TEXTURE_3D_EXT);
		//glDisable(GL_BLEND);

	glPopMatrix(); // back to identity

    glMatrixMode(GL_MODELVIEW);

//	printf("5: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);

	//raz
	cgGLDisableProfile( vertexProfile );
	cgGLDisableProfile(fragmentProfile);

	glActiveTextureARB(GL_TEXTURE2_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_3D_EXT);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glDisable(GL_TEXTURE_3D_EXT);
	glDisable(GL_TEXTURE_2D);

	
    glDisable(GL_CLIP_PLANE0);
    glDisable(GL_CLIP_PLANE1);
    glDisable(GL_CLIP_PLANE2);
    glDisable(GL_CLIP_PLANE3);
    glDisable(GL_CLIP_PLANE4);
    glDisable(GL_CLIP_PLANE5);


//	printf("6: %d\n", glutGet(GLUT_ELAPSED_TIME)-ttt);
	/*glPushMatrix();
		glScalef(90,90,90);
		glRotatef(objangle[X], 0.f, 0.f, 1.f);
		glRotatef(objangle[Y], -1.f, 0.f, 0.f);

		drawCube();
	glPopMatrix();*/




}

void VolRender::buildPlanes(){
	GLfloat offS, offT, offR; // mapping texture to planes 

    offS = 100.f/fieldSize.x;
    offT = 100.f/fieldSize.y;
    offR = 100.f/fieldSize.z;


	glNewList(planesList, GL_COMPILE);


	for(int i = 0; i < slices; i++) {
		glBegin(GL_QUADS);

		glTexCoord3f(0, 0, (1-(float)i/(float)slices));
		glVertex3f(-100.f, -100.f, 
		   -100.f + offR + i * (200.f - 2 * offR)/(slices - 1)); //200.f-2

		glTexCoord3f(1, 0, 1-(float)i/(float)slices);
		glVertex3f( 100.f, -100.f,
		   -100.f + offR + i * (200.f - 2 * offR)/(slices - 1));

		glTexCoord3f(1, 1, 1-(float)i/(float)slices);
		glVertex3f( 100.f,  100.f,
		   -100.f + offR + i * (200.f - 2 * offR)/(slices - 1));

		glTexCoord3f(0, 1, 1-(float)i/(float)slices);
		glVertex3f(-100.f,  100.f,
		   -100.f + offR + i * (200.f - 2 * offR)/(slices - 1));

		glEnd();
	}

	glEndList();
}

//Calcul des gradients pour l'clairage
void VolRender::computeGradient(){

	int tx=fieldSize.x, ty=fieldSize.y, tz=fieldSize.z;
	int nbcomp=3;

	if(gradField)
		delete [] gradField;

	gradField=new float[fieldSize.x*fieldSize.y*fieldSize.z*nbcomp];


	for(int k=0; k<tz-0; k++)
		for(int j=0; j<ty-0; j++)
			for(int i=0; i<tx-0; i++){
			/*	gradField[0 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=(float)i/(float)tx;//tex3ddata[0 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty];//(tex3ddata[0 + (i+1)*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty] - tex3ddata[0 + (i-1)*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]);
				gradField[1 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=(float)j/(float)ty;//tex3ddata[1 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty];//(tex3ddata[0 + i*nbcomp + (j+1)*nbcomp*tx + k*nbcomp*tx*ty] - tex3ddata[0 + i*nbcomp + (j-1)*nbcomp*tx + k*nbcomp*tx*ty]);
				gradField[2 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=(float)k/(float)tz;//tex3ddata[2 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty];//(tex3ddata[0 + i*nbcomp + j*nbcomp*tx + (k+1)*nbcomp*tx*ty] - tex3ddata[0 + i*nbcomp + j*nbcomp*tx + (k-1)*nbcomp*tx*ty]);
			*/
			/*	gradField[0 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=tex3ddata[0 + i*1 + j*1*tx + k*1*tx*ty];//(tex3ddata[0 + (i+1)*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty] - tex3ddata[0 + (i-1)*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]);
				gradField[1 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=tex3ddata[0 + i*1 + j*1*tx + k*1*tx*ty];//(tex3ddata[0 + i*nbcomp + (j+1)*nbcomp*tx + k*nbcomp*tx*ty] - tex3ddata[0 + i*nbcomp + (j-1)*nbcomp*tx + k*nbcomp*tx*ty]);
				gradField[2 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=tex3ddata[0 + i*1 + j*1*tx + k*1*tx*ty];//(tex3ddata[0 + i*nbcomp + j*nbcomp*tx + (k+1)*nbcomp*tx*ty] - tex3ddata[0 + i*nbcomp + j*nbcomp*tx + (k-1)*nbcomp*tx*ty]);
			*/
				
			//	int di=4, dj=4, dk=4;

				if(i<1 || j<1 || k<1 || i>(tx-2) || j>(ty-2) || k>(tz-2)){
					gradField[0 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=0;
					gradField[1 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=0;
					gradField[2 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=0;
				}else{

					int di, dj, dk;

					if(i<5)
						di=i-1;
					else if(i>tx-6)
							di=-2-i+tx;
						 else
							 di=4;

					if(j<5)
						dj=j-1;
					else if(j>ty-6)
							dj=-2-j+ty;
						 else
							 dj=4;

					if(k<5)
						dk=k-1;
					else if(k>tz-6)
							dk=-2-k+tz;
						 else
							 dk=4;

					gradField[0 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=(tex3ddata[ (i+di) + j*tx + k*tx*ty] - tex3ddata[ (i-di) + j*tx + k*tx*ty])/2.0 +0.5;
					gradField[1 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=(tex3ddata[ i + (j+dj)*tx + k*tx*ty] - tex3ddata[ i + (j-dj)*tx + k*tx*ty])/2.0 +0.5;
					gradField[2 + i*nbcomp + j*nbcomp*tx + k*nbcomp*tx*ty]=(tex3ddata[ i + j*tx + (k+dk)*tx*ty] - tex3ddata[ i + j*tx + (k-dk)*tx*ty])/2.0 +0.5;
				}
			}

	glBindTexture(GL_TEXTURE_3D_EXT, gradID);

	gluBuild3DMipmaps(GL_TEXTURE_3D_EXT, 3, tx, ty, tz, GL_RGB, GL_FLOAT, gradField);
//	glTexImage3DEXT(GL_TEXTURE_3D_EXT, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, tx, ty, tz, 0, GL_RGB , GL_FLOAT, gradField);

	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_CLAMP);

	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

}


void VolRender::drawCube(){

	glDisable(GL_TEXTURE_2D);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	
	//glBlendFunc(GL_ONE, GL_ONE_MINUS_DST_ALPHA);
	//glBlendFunc(GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR);

	glColor4f ( 1, 1, 1, 0.5);

		glBegin(GL_QUADS);
		// Front Face
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);	// Top Right Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);	// Top Left Of The Texture and Quad
		// Back Face
		glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);	// Top Right Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);	// Top Left Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);	// Bottom Left Of The Texture and Quad
		// Top Face
		glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);	// Top Left Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);	// Top Right Of The Texture and Quad
		// Bottom Face
		glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f, -1.0f, -1.0f);	// Top Right Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f, -1.0f, -1.0f);	// Top Left Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
		// Right face
		glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);	// Top Right Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);	// Top Left Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);	// Bottom Left Of The Texture and Quad
		// Left Face
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);	// Top Right Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);	// Top Left Of The Texture and Quad
	glEnd();
	
	glEnable(GL_TEXTURE_2D);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

}
